import os
from pathlib import Path
from typing import Union, Optional, Any, Dict, TypeVar

from owlsight.hugging_face.constants import HUGGINGFACE_TASKS

# ANSI color codes for terminal output
COLOR_CODES = {
    "red": "\033[31m",
    "green": "\033[32m",
    "yellow": "\033[33m",
    "blue": "\033[34m",
    "magenta": "\033[35m",
    "cyan": "\033[36m",
    "reset": "\033[0m",  # Resets to default color
}

PROMPT_COLOR = "blue"
CHOICE_COLOR = "green"

T = TypeVar("T")


class ConfigSchema:
    """Configuration schema definition and validation."""

    SCHEMA = {
        "main": {
            "max_retries_on_error": {
                "default": 3,
                "choices": list(range(0, 10)),
                "description": "Maximum number of retries for error recovery",
            },
            "prompt_retry_on_error": {
                "default": False,
                "choices": [False, True],
                "description": "Whether to prompt before retrying on error",
            },
            "prompt_code_execution": {
                "default": True,
                "choices": [False, True],
                "description": "Whether to prompt before executing code",
            },
            "extra_index_url": {
                "default": "",
                "choices": None,
                "description": "Additional URL for package installation",
            },
        },
        "model": {
            "model_id": {"default": "", "choices": None, "description": "Model identifier or path"},
            "save_history": {
                "default": False,
                "choices": [False, True],
                "description": "Whether to save conversation history",
            },
            "system_prompt": {
                "default": "",
                "choices": None,
                "description": "System prompt defining model behavior",
            },
            # Transformers specific
            "transformers__device": {
                "default": None,
                "choices": [None, "cpu", "cuda", "mps"],
                "description": "Device for transformers model",
            },
            "transformers__quantization_bits": {
                "default": None,
                "choices": [None, 8, 4],
                "description": "Quantization bits for transformers model",
            },
            "transformers__stream": {
                "default": True,
                "choices": [False, True],
                "description": "Whether to stream input to transformers model",
            },
            "transformers__use_fp16": {
                "default": False,
                "choices": [False, True],
                "description": "Whether to use FP16 for transformers model",
            },
            # GGUF specific
            "gguf__filename": {"default": "", "choices": None, "description": "GGUF model filename"},
            "gguf__verbose": {
                "default": False,
                "choices": [False, True],
                "description": "Verbose output for GGUF model",
            },
            "gguf__n_ctx": {
                "default": 512,
                "choices": [32 * (2**n) for n in range(15)],
                "description": "Context length for GGUF model",
            },
            "gguf__n_gpu_layers": {
                "default": 0,
                "choices": [-1, 0, 1] + [(2**n) for n in range(1, 9)],
                "description": "Number of GPU layers for GGUF model",
            },
            "gguf__n_batch": {
                "default": 512,
                "choices": [32 * (2**n) for n in range(11)],
                "description": "Batch size for GGUF model",
            },
            "gguf__n_cpu_threads": {
                "default": 1,
                "choices": list(range(1, os.cpu_count() + 1)),
                "description": "Number of CPU threads for GGUF model",
            },
            # ONNX specific
            "onnx__tokenizer": {"default": "", "choices": None, "description": "Tokenizer for ONNX model"},
            "onnx__verbose": {
                "default": False,
                "choices": [False, True],
                "description": "Verbose output for ONNX model",
            },
            "onnx__num_threads": {
                "default": 1,
                "choices": list(range(1, os.cpu_count() + 1)),
                "description": "Number of threads for ONNX model",
            },
        },
        "generate": {
            "stopwords": {"default": [], "choices": None, "description": "Words that stop text generation"},
            "max_new_tokens": {
                "default": 512,
                "choices": [32 * (2**n) for n in range(15)],
                "description": "Maximum tokens to generate",
            },
            "temperature": {
                "default": 0.0,
                "choices": [round(x * 0.05, 2) for x in range(21)],
                "description": "Temperature for text generation",
            },
            "generation_kwargs": {"default": {}, "choices": None, "description": "Additional generation parameters"},
        },
        "rag": {
            "active": {"default": False, "choices": [False, True], "description": "Whether RAG is active"},
            "target_library": {"default": "", "choices": None, "description": "Target python library for RAG"},
            "top_k": {"default": 3, "choices": list(range(1, 51)), "description": "Number of RAG results to return"},
            "search_query": {"default": "", "choices": None, "description": "RAG search query"},
        },
        "huggingface": {
            "search": {"default": "", "choices": None, "description": "search for a model on huggingface"},
            "top_k": {
                "default": 5,
                "choices": list(range(1, 51)),
                "description": "Number of huggingface results to return",
            },
            "select_model": {"default": "", "choices": None, "description": "select a model from huggingface"},
            "task": {
                "default": None,
                "choices": HUGGINGFACE_TASKS,
                "description": "task for huggingface",
            },
        },
    }

    @classmethod
    def get_defaults(cls) -> Dict[str, Dict[str, Any]]:
        """Extract default values from schema."""
        return {
            section: {key: value["default"] for key, value in options.items()}
            for section, options in cls.SCHEMA.items()
        }

    @classmethod
    def get_choices(cls) -> Dict[str, Dict[str, Any]]:
        """Extract choices from schema, adding 'back' option to each section."""
        return {
            section: {
                "back": None,
                **{
                    key: value["choices"] if value["choices"] is not None else value["default"]
                    for key, value in options.items()
                },
            }
            for section, options in cls.SCHEMA.items()
        }

    @classmethod
    def get_descriptions(cls) -> Dict[str, Dict[str, str]]:
        """Extract descriptions from schema."""
        return {
            section: {key: value["description"] for key, value in options.items()}
            for section, options in cls.SCHEMA.items()
        }

    # @classmethod
    # def validate_value(cls, section: str, key: str, value: Any) -> bool:
    #     """Validate a configuration value against the schema."""
    #     if section not in cls.SCHEMA or key not in cls.SCHEMA[section]:
    #         return False

    #     choices = cls.SCHEMA[section][key]["choices"]
    #     if choices is None:
    #         return isinstance(value, type(cls.SCHEMA[section][key]["default"]))
    #     return value in choices

    # @classmethod
    # def get_next_value(cls, section: str, key: str, current_value: T) -> T:
    #     """Get next value from choices, cycling back to first if at end."""
    #     choices = cls.SCHEMA[section][key]["choices"]
    #     if choices is None or not choices:
    #         return current_value

    #     try:
    #         current_idx = choices.index(current_value)
    #         next_idx = (current_idx + 1) % len(choices)
    #         return choices[next_idx]
    #     except ValueError:
    #         return choices[0] if choices else current_value

    # @classmethod
    # def get_prev_value(cls, section: str, key: str, current_value: T) -> T:
    #     """Get previous value from choices, cycling to last if at start."""
    #     choices = cls.SCHEMA[section][key]["choices"]
    #     if choices is None or not choices:
    #         return current_value

    #     try:
    #         current_idx = choices.index(current_value)
    #         prev_idx = (current_idx - 1) % len(choices)
    #         return choices[prev_idx]
    #     except ValueError:
    #         return choices[-1] if choices else current_value


# Create global constants from schema
DEFAULTS = ConfigSchema.get_defaults()
CHOICES = ConfigSchema.get_choices()
DESCRIPTIONS = ConfigSchema.get_descriptions()

# Menu configuration
MENU_KEYS = {
    "assistant": "how can I assist you?",
}

MAIN_MENU = {
    MENU_KEYS["assistant"]: "",
    "shell": "",
    "python": None,
    "config": list(DEFAULTS.keys()),
    "save": "",
    "load": "",
    "clear history": None,
    "quit": None,
}


# Directory and file handling functions
def get_cache_dir() -> Path:
    """Returns the base directory for storing cached data."""
    data_dir = Path.home() / ".owlsight"
    data_dir.mkdir(parents=True, exist_ok=True)
    return data_dir


def create_directory(path: Union[str, Path], base: Optional[Path] = None) -> Path:
    """Creates a directory if it does not exist and returns the path."""
    full_path = Path(base or get_cache_dir()) / path
    full_path.mkdir(parents=True, exist_ok=True)
    return full_path


def create_file(path: Union[str, Path], base: Optional[Path] = None) -> Path:
    """Creates an empty file if it does not exist and returns the file path."""
    full_path = Path(base or get_cache_dir()) / path
    full_path.parent.mkdir(parents=True, exist_ok=True)
    full_path.touch(exist_ok=True)
    return full_path


def get_prompt_cache() -> Path:
    """Returns the path to the prompt history cache file."""
    return create_file(".prompt_history")


def get_py_cache() -> Path:
    """Returns the path to the python history cache file."""
    return create_file(".python_history")


def get_pickle_cache() -> Path:
    """Returns the path to the pickle cache directory."""
    return create_directory(".pickle")



from pathlib import Path
import requests
import io
import ast
import sys
import pytesseract

sys.path.append("src")

from PIL import Image
import numpy as np
import pytest

from owlsight.hugging_face.constants import HUGGINGFACE_MEDIA_TASKS
from owlsight.processors.multimodal_processors import MultiModalProcessorTransformers
from owlsight.utils.custom_classes import MediaObject
from owlsight.multimodal.tesseract import find_tesseract_installation
pytesseract.pytesseract.tesseract_cmd = find_tesseract_installation()


# Test URLs
TEST_CASES = [
    {
        "task": "image-to-text",
        "url": "https://upload.wikimedia.org/wikipedia/commons/d/d3/Statue_of_Liberty%2C_NY.jpg",
        "question": None,
        "expected_type": list,
        "media_type": "image",
    },
    {
        "task": "visual-question-answering",
        "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/9/96/VW_K%C3%A4fer_Baujahr_1966.jpg/420px-VW_K%C3%A4fer_Baujahr_1966.jpg",
        "question": "What color is the car?",
        "expected_type": list,
        "media_type": "image",
    },
    {
        "task": "automatic-speech-recognition",
        "url": "https://www2.cs.uic.edu/~i101/SoundFiles/gettysburg10.wav",
        "question": None,
        "expected_type": dict,
        "media_type": "audio",
    },
    {
        "task": "document-question-answering",
        "url": "https://vt-vtwa-assets.varsitytutors.com/vt-vtwa/uploads/problem_question_image/image/19791/table.jpg",
        "question": "What is the total number of students?",
        "expected_type": list,
        "media_type": "image",
    },
]


@pytest.fixture(scope="module")
def test_data():
    """Download and cache test data."""
    cached_data = {}
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3"
    }
    for case in TEST_CASES:
        response = requests.get(case["url"], headers=headers, timeout=10)
        response.raise_for_status()
        cached_data[case["task"]] = response.content
    return cached_data


@pytest.fixture(params=TEST_CASES)
def processor(request, media_model_mappings):
    """Create processor for each test case."""
    return MultiModalProcessorTransformers(
        model_id=media_model_mappings[request.param["task"]], task=request.param["task"]
    )


def test_media_preprocessor_initialization(media_model_mappings):
    """Test MediaPreprocessor initialization with various tasks."""
    for task in HUGGINGFACE_MEDIA_TASKS:
        processor = MultiModalProcessorTransformers(model_id=media_model_mappings[task], task=task)
        assert processor.task == task
        assert processor.media_preprocessor is not None
        assert processor.text_processor.pipe is not None


def test_invalid_task():
    """Test initialization with invalid task."""
    with pytest.raises(ValueError):
        MultiModalProcessorTransformers(model_id="test", task="invalid_task")


@pytest.mark.parametrize("case", TEST_CASES)
def test_generate(case, test_data, media_model_mappings):
    """Test generate method for each task."""
    processor = MultiModalProcessorTransformers(model_id=media_model_mappings[case["task"]], task=case["task"])

    # Create temporary file with test data
    test_file = Path(f"test_{case['task']}_file.tmp")
    test_file.write_bytes(test_data[case["task"]])

    try:
        # Create MediaObject for the test case
        media_objects = {"__MEDIA_0__": MediaObject(type=case["media_type"], path=str(test_file), options={})}

        # Prepare input text based on whether there's a question
        input_text = f"{case['question']} __MEDIA_0__" if case["question"] else "__MEDIA_0__"

        result = processor.generate(input_text, media_objects=media_objects)
        result = ast.literal_eval(result)

        assert isinstance(result, case["expected_type"])
        assert len(result) > 0

    finally:
        # Clean up temporary file
        if test_file.exists():
            test_file.unlink()


def test_preprocessing(media_model_mappings):
    """Test preprocessing for different input types."""
    processor = MultiModalProcessorTransformers(model_id=media_model_mappings["image-to-text"], task="image-to-text")

    # Create test image
    test_image = Image.new("RGB", (100, 100), color="red")
    buffer = io.BytesIO()
    test_image.save(buffer, format="PNG")

    # Test with bytes
    result = processor.media_preprocessor.preprocess_input(buffer.getvalue())
    assert isinstance(result, Image.Image)

    # Test with Path
    test_image.save("test_image.png")
    result = processor.media_preprocessor.preprocess_input(Path("test_image.png"))
    assert isinstance(result, Image.Image)
    Path("test_image.png").unlink()


def test_audio_preprocessing(test_data, media_model_mappings):
    """Test audio preprocessing specifically."""
    processor = MultiModalProcessorTransformers(
        model_id=media_model_mappings["automatic-speech-recognition"], task="automatic-speech-recognition"
    )

    result = processor.media_preprocessor.preprocess_input(test_data["automatic-speech-recognition"])

    assert "array" in result
    assert "sampling_rate" in result
    assert isinstance(result["array"], np.ndarray)
    assert result["sampling_rate"] == 16000


def test_error_handling(media_model_mappings):
    """Test error handling for invalid inputs."""
    processor = MultiModalProcessorTransformers(model_id=media_model_mappings["image-to-text"], task="image-to-text")

    # Test with non-existent file
    media_objects = {"__MEDIA_0__": MediaObject(type="image", path="non_existent_file.jpg", options={})}

    with pytest.raises(FileNotFoundError):
        processor.generate("__MEDIA_0__", media_objects=media_objects)

    # Test with invalid URL
    media_objects = {"__MEDIA_0__": MediaObject(type="image", path="https://invalid.url/image.jpg", options={})}

    with pytest.raises(requests.exceptions.RequestException):
        processor.generate("__MEDIA_0__", media_objects=media_objects)


if __name__ == "__main__":
    pytest.main(["-vvv", __file__])



import pytest
import sys
import tempfile
import os

sys.path.append("src")
sys.path.append("tests")
from owlsight.utils.code_execution import CodeExecutor, execute_code_with_feedback
from owlsight.utils.venv_manager import get_pyenv_path, get_pip_path


@pytest.fixture
def code_executor(text_generation_manager):
    return CodeExecutor(
        text_generation_manager,
        temp_dir="temp_dir",
        pip_path="pip",
        pyenv_path="venv",
    )


def test_code_executor_execute_python_code_succesfully(code_executor: CodeExecutor):
    result = code_executor.execute_and_retry("python", "print('Hello')", "original question")
    assert result


def test_code_executor_python_state_is_saved(code_executor: CodeExecutor):
    # Test that variables are correctly set in the state after execution
    code_executor.execute_python_code("x = 10")
    assert code_executor.globals_dict.get("x") == 10

    # Test executing another block of code that uses the state
    code_executor.execute_python_code("y = x + 5")
    assert code_executor.globals_dict.get("y") == 15

    # Test that the state is persistent across executions
    assert code_executor.globals_dict.get("x") == 10  # x should still be available


def test_clear_state(code_executor: CodeExecutor):
    # Set some state
    code_executor.execute_python_code("x = 20")
    assert code_executor.globals_dict.get("x") == 20

    # Clear the state and check if it's removed
    code_executor.globals_dict.clear()
    assert code_executor.globals_dict.get("x") is None  # State should be cleared


def test_code_executor_install_missing_module_in_venv(text_generation_manager):
    # arrange
    module_name = "tinydb"
    question = f"Use python to generate code which uses the '{module_name}' module"
    model_response = f"""
    Sure! Here is a Python code snippet that imports the '{module_name}' module:
    ```python
    import {module_name} as md\na = 5
    ```
    """.strip()
    with tempfile.TemporaryDirectory() as temp_dir:
        pyenv_path = get_pyenv_path()
        pip_path = get_pip_path(pyenv_path)

        code_executor = CodeExecutor(text_generation_manager, pyenv_path, pip_path, temp_dir)
        results = execute_code_with_feedback(
            model_response,
            question,
            code_executor,
            prompt_code_execution=False,
        )

        assert isinstance(results, list)
        assert len(results) == 1
        assert isinstance(results[0], dict)
        assert results[0]["success"]

        # lib is installed in venv
        assert module_name in os.listdir(code_executor.temp_dir)

        # state is saved correctly
        assert code_executor.globals_dict.get("a") == 5



"""helper functions for huggingface models"""

import os
import numpy as np
from typing import Iterable, Optional, Union, Dict, List
import requests
import subprocess
from datetime import datetime, timezone
from functools import lru_cache

from huggingface_hub import HfApi
from huggingface_hub.hf_api import ModelInfo
from tqdm import tqdm

from owlsight.utils.helper_functions import check_invalid_input_parameters
from owlsight.utils.logger import logger

MODELHUB_PREFIX = "https://huggingface.co/"


@lru_cache(maxsize=128)
def calculate_days_ago_created(time_created: datetime) -> int:
    """
    Calculate the number of days since a given time.
    """
    now_utc = datetime.now(timezone.utc)
    time_difference = now_utc - time_created
    days_difference = time_difference.days

    return days_difference


def engagement_score(likes: int, downloads: int, created_days_ago: int) -> float:
    """
    Calculate an engagement score based on likes, downloads and days since creation.

    This function computes an engagement score that takes into account
    the number of likes, downloads and days since creation for a model, and the number of days ago a model was created.
    It applies stronger penalties for very low engagement (low likes and downloads),
    and uses logarithmic scaling to balance disparities between likes and downloads.

    Parameters
    ----------
    likes : int
        The number of likes for the model.
    downloads : int
        The number of downloads for the model.
    created_days_ago : int
        The number of days since the model was created.

    Returns
    -------
    float
        The calculated engagement score.
    """
    # Constants
    C: int = 10  # Offset to avoid zero division
    download_threshold = 50  # Threshold below which downloads are heavily penalized

    # Base score calculation with penalty for very low downloads and likes
    if likes == 0 and downloads == 0:
        base_score = 0  # No engagement, so score should be zero
    else:
        # Logarithmic scaling for likes and downloads to reduce impact of low values
        base_score = (np.log1p(likes) / (np.log1p(downloads + C))) ** 1.2

    # Download factor with clamping and penalty for low downloads
    download_factor: float = np.log(downloads + C) / np.log(1000)
    download_factor = min(max(download_factor, 0.1), 1)  # Clamp between 0.1 and 1

    # Penalize very low downloads
    low_download_penalty = 0.2 if downloads < download_threshold else 1

    # Adjustment for zero likes
    zero_likes_penalty: float = 0.5 if likes == 0 else 1

    days_ago_score = calculate_days_ago_score(created_days_ago)

    # Final score calculation
    score: float = (base_score * download_factor * zero_likes_penalty * low_download_penalty) + days_ago_score

    return score


def _make_hashable(value):
    """Convert potentially unhashable types to hashable ones for caching."""
    if isinstance(value, (list, set)):
        return tuple(sorted(_make_hashable(x) for x in value))
    if isinstance(value, dict):
        return tuple(sorted((k, _make_hashable(v)) for k, v in value.items()))
    return value


@lru_cache(maxsize=128)
def _cached_get_model_list(
    filter_by: Optional[tuple] = None, search: Optional[str] = None, include_metadata: bool = False, **kwargs
) -> tuple:
    """Cached inner function that returns all models without top_n slicing."""
    hf_api = HfApi()

    # Convert filter_by back to list if it's not None
    filter_by_list = list(filter_by) if filter_by is not None else None

    check_invalid_input_parameters(hf_api.list_models, kwargs)
    model_gen = hf_api.list_models(
        filter=filter_by_list,
        search=search,
        cardData=include_metadata,
        **kwargs,
    )

    model_info_list = []

    for model_info in model_gen:
        days_ago_created = calculate_days_ago_created(model_info.created_at)
        score = engagement_score(model_info.likes, model_info.downloads, days_ago_created)
        model_info.engagement_score = score
        model_info_list.append(model_info)

    model_info_list.sort(key=lambda x: x.engagement_score, reverse=True)

    # Convert to tuple for caching
    return tuple(model_info_list)


def get_model_list(
    filter_by: Union[str, Iterable[str], None] = None,
    top_n: Optional[int] = None,
    include_metadata: bool = False,
    search: Optional[str] = None,
    **kwargs,
) -> List[ModelInfo]:
    """
    Get a list of models that match the given criteria, with caching support.

    Parameters
    ----------
    filter_by : Union[str, Iterable[str], None], optional
        A string or list of strings to filter the models by. Defaults to None.
    top_n : int, optional
        The number of models to return. Defaults to None.
    include_metadata : bool, optional
        Whether to add extra metadata to the model. Defaults to False.
    search : str, optional
        A string to filter the models by during searching. Defaults to None.
    hf_api : HfApi, optional
        An instance of HuggingFace API to use. If None, a new instance will be created. Defaults to None.

    Returns
    -------
    List[ModelInfo]
        A list of models that match the given criteria.
    """
    # Convert filter_by to hashable type (tuple) for caching
    hashable_filter_by = _make_hashable(filter_by) if filter_by is not None else None

    # Convert kwargs to hashable format
    hashable_kwargs = _make_hashable(kwargs)

    # Get full cached result
    cached_results = _cached_get_model_list(
        filter_by=hashable_filter_by, search=search, include_metadata=include_metadata, **dict(hashable_kwargs)
    )

    # Convert back to list and apply top_n if specified
    model_info_list = list(cached_results)
    if top_n:
        model_info_list = model_info_list[:top_n]

    return model_info_list


def download_huggingface_model(model_name: str, save_path: str, chunk_size: int = 1024) -> None:
    """
    Construct the URL to download the model

    Parameters
    ----------
    model_name : str
        The name of the model to download
    save_path : str
        The path where the model will be saved
    chunk_size : int
        The size of each chunk to download

    Returns
    -------
    None
    """
    base_url = f"https://huggingface.co/{model_name}/resolve/main/"
    file_names = [
        "config.json",
        "pytorch_model.bin",
        "tokenizer_config.json",
        "vocab.txt",
        "special_tokens_map.json",
    ]

    # Create the folder where the model will be saved
    os.makedirs(save_path, exist_ok=True)

    for file_name in file_names:
        # Download each file
        url = base_url + file_name
        response = requests.get(url, stream=True)

        if response.status_code == 200:
            # Show progress bar
            total_size_in_bytes = int(response.headers.get("content-length", 0))
            progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)

            with open(os.path.join(save_path, file_name), "wb") as f:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    progress_bar.update(len(chunk))
                    f.write(chunk)
            progress_bar.close()
        else:
            logger.error("Failed to download %s due to %s", url, response.status_code)


def show_model_memory(model_name: str) -> Optional[str]:
    """
    Executes a shell command to estimate the memory usage of a given model.

    Args:
        model_name (str): Name of the model to estimate memory for.

    Returns:
        Optional[str]: The output of the memory estimation command, or None if an error occurs.
    """
    command = [
        "accelerate",
        "estimate-memory",
        model_name,
        "--library_name",
        "transformers",
    ]

    try:
        # Execute the command and capture the output
        result = subprocess.run(
            command,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            encoding="utf-8",  # Specify UTF-8 encoding
            errors="ignore",  # Ignore decoding errors
        )
        return result.stdout
    except subprocess.CalledProcessError as e:
        print(f"An error occurred while estimating memory: {e.stderr}")
        return None
    except UnicodeDecodeError as e:
        print(f"A Unicode decoding error occurred: {e}")
        return None


def _get_hf_model_data(model_info: "ModelInfo") -> Dict[str, str]:
    days_difference = calculate_days_ago_created(model_info.created_at)
    model_data = {
        "engagement score": np.round(model_info.engagement_score, 4),
        "days ago created": days_difference,
        "downloads": model_info.downloads,
        "likes": model_info.likes,
        "url": os.path.join(MODELHUB_PREFIX, model_info.modelId),
    }

    return model_data


def calculate_days_ago_score(created_days_ago, weight=0.01, N=0.3):
    """
    Optimized version with better score distribution:
    Reduced weight to prevent scores from decaying too quickly

    Parameters
    ----------
    created_days_ago : int
        The number of days since the model was created.
    weight : float, optional
        The weight factor for the exponential decay. Defaults to 0.01.
    N : float, optional
        Increase this value to boost the score. Defaults to 0.25.
    """
    if not created_days_ago:
        return 0.0

    return np.exp(-(weight * (1 + created_days_ago))) * N



from typing import Optional, Type
import os
import traceback

from transformers.pipelines import get_task

from owlsight.hugging_face.constants import HUGGINGFACE_MEDIA_TASKS
from owlsight.processors.base import TextGenerationProcessor
from owlsight.processors.multimodal_processors import MultiModalProcessorTransformers
from owlsight.processors.text_generation_processors import (
    TextGenerationProcessorGGUF,
    TextGenerationProcessorOnnx,
    TextGenerationProcessorTransformers,
)
from owlsight.utils.logger import logger


def _select_transformers_processor_type_on_task(
    model_id: str,
    task: Optional[str] = None,
) -> TextGenerationProcessor:
    if not task:
        try:
            task = get_task(model_id)
        except Exception as e:
            logger.error(
                f"Error while trying to infer the task for model {model_id}: {e}. Inferred TextGenerationProcessor might not be the correct one."
            )

    if task in HUGGINGFACE_MEDIA_TASKS:
        return MultiModalProcessorTransformers

    return TextGenerationProcessorTransformers


def select_processor_type(model_id: str, task: Optional[str] = None) -> Type["TextGenerationProcessor"]:
    """
    Utilityfunction which selects the appropriate TextGenerationProcessor class based on the model ID or directory.

    If the model_id is a directory, the function will inspect the contents of the directory
    to decide the processor type. Otherwise, it will use the model_id string to make the decision.
    """
    # Check if the model_id is a directory
    if os.path.isdir(model_id):
        # Check if any file in the directory ends with .onnx
        if any(f.endswith("onnx") for f in os.listdir(model_id)):
            return TextGenerationProcessorOnnx
        elif model_id.lower().endswith("gguf") or any(f.endswith("gguf") for f in os.listdir(model_id)):
            return TextGenerationProcessorGGUF
        else:
            return _select_transformers_processor_type_on_task(model_id, task)
    else:
        # If model_id is not a directory, use the model_id string
        if model_id.lower().endswith("gguf"):
            return TextGenerationProcessorGGUF
        elif "onnx" in model_id.lower():
            return TextGenerationProcessorOnnx
        else:
            return _select_transformers_processor_type_on_task(model_id, task)



import pytest
from owlsight.ui.console import OptionType, Selector


@pytest.fixture
def options_dict():
    """
    Fixture that provides a standard dictionary with editable input,
    static options, and a toggle option.
    """
    return {
        "You are a:": "",  # Editable input
        "apple": None,  # Static option
        "pear": None,  # Static option
        "Is it ripe?": [True, False],  # Toggleable option
    }


def test_selector_initialization(options_dict):
    """
    Test that the Selector class correctly initializes with options.
    """
    selector = Selector(options_dict)

    # Ensure that the options are properly parsed and initialized
    assert len(selector.options) == 4
    assert selector.options[0] == ("You are a:", OptionType.EDITABLE)
    assert selector.options[1] == ("apple", OptionType.SINGLE)
    assert selector.options[2] == ("pear", OptionType.SINGLE)
    assert selector.options[3] == ("Is it ripe?", OptionType.TOGGLE)

    # Ensure that the initial values of user inputs and toggles are correct
    assert selector.user_inputs == {"You are a:": ""}
    assert selector.toggle_values == {"Is it ripe?": True}
    assert selector.current_index == 0



import sys
sys.path.append("src")

import pytest
from transformers import AutoTokenizer
from owlsight.processors.text_generation_processors import TextGenerationProcessorGGUF


@pytest.fixture
def setup_gguf_processor():
    """Fixture to set up the GGUF processor and tokenizer."""
    model_id = "TheBloke/TinyLlama-1.1B-Chat-v1.0-GGUF"
    gguf_filename = "tinyllama-1.1b-chat-v1.0.Q2_K.gguf"
    tokenizer = AutoTokenizer.from_pretrained("TinyLlama/TinyLlama-1.1B-Chat-v1.0")
    processor = TextGenerationProcessorGGUF(model_id, gguf__filename=gguf_filename)
    return processor, tokenizer


def test_gguf_generate_response(setup_gguf_processor):
    """Test that the GGUF processor generates a valid response."""
    processor, _ = setup_gguf_processor
    prompt = "test prompt"
    max_new_tokens = 128

    response = processor.generate(prompt, max_new_tokens=max_new_tokens)
    assert isinstance(response, str), "Generated response should be a string."


def test_gguf_token_count_within_tolerance(setup_gguf_processor):
    """Test that the GGUF generated token count is within the acceptable range."""
    processor, tokenizer = setup_gguf_processor
    prompt = "test prompt"
    max_new_tokens = 128
    tolerance_fraction = 0.5

    response = processor.generate(prompt, max_new_tokens=max_new_tokens)

    # Tokenize response and calculate new tokens
    response_tokens = tokenizer.tokenize(response)
    prompt_tokens = tokenizer.tokenize(prompt)
    new_tokens = response_tokens[len(prompt_tokens):]

    # Calculate acceptable range
    lower_bound = max_new_tokens - (max_new_tokens * tolerance_fraction)
    upper_bound = max_new_tokens + (max_new_tokens * tolerance_fraction)

    assert lower_bound <= len(new_tokens) <= upper_bound, (
        f"Expected approximately {max_new_tokens} tokens "
        f"(range: {lower_bound:.2f} - {upper_bound:.2f}), got {len(new_tokens)}"
    )


def test_gguf_prompt_tokens_exclusion(setup_gguf_processor):
    """Test that the GGUF response excludes prompt tokens when counting new tokens."""
    processor, tokenizer = setup_gguf_processor
    prompt = "test prompt"
    max_new_tokens = 128

    response = processor.generate(prompt, max_new_tokens=max_new_tokens)

    # Tokenize response and calculate new tokens
    response_tokens = tokenizer.tokenize(response)
    prompt_tokens = tokenizer.tokenize(prompt)

    assert len(response_tokens) > len(prompt_tokens), "Response tokens should exceed prompt tokens."


if __name__ == "__main__":
    pytest.main([__file__])



from typing import Dict, List, Optional
from .helper_functions import get_model_list, _get_hf_model_data


def get_model_data(model_search: str, top_n_models: int = 10, **kwargs) -> Dict[str, Dict[str, str]]:
    """
    Get and display the model data from the HuggingFace Hub in a visually appealing format.

    Parameters:
        model_search: Search term for filtering models
        top_n_models: Number of top models to display
        **kwargs: Additional keyword arguments to pass to get_model_list. E.g., task, framework, etc.
        See `HfApi().list_models()` from `huggingface_hub` package for more details.

    Returns:
        Dictionary containing model information
    """
    model_list = get_model_list(top_n=top_n_models, search=model_search, **kwargs)
    model_dict = {model_info.modelId: _get_hf_model_data(model_info) for model_info in model_list}
    return model_dict


def create_separator(width: int = 80, char: str = "=") -> str:
    """Create a separator line with given width and character."""
    return char * width


def format_table_row(columns: List[str], widths: List[int]) -> str:
    """Format a row with proper spacing and alignment."""
    return "│ " + " │ ".join(col.ljust(width) for col, width in zip(columns, widths)) + " │"


def calculate_column_widths(model_dict: Dict[str, Dict[str, str]]) -> tuple:
    """Calculate optimal column widths based on content."""
    max_model_id = max(len(model_id) for model_id in model_dict.keys())
    max_detail = max(
        len(f"{key}: {str(value)}") for model_data in model_dict.values() for key, value in model_data.items()
    )

    RANK_WIDTH = 4
    MODEL_ID_WIDTH = max(max_model_id + 2, 30)  # minimum 30 chars
    DETAIL_WIDTH = max(max_detail + 2, 40)  # minimum 40 chars
    TOTAL_WIDTH = RANK_WIDTH + MODEL_ID_WIDTH + DETAIL_WIDTH + 10  # padding and separators

    return RANK_WIDTH, MODEL_ID_WIDTH, DETAIL_WIDTH, TOTAL_WIDTH


def show_and_return_model_data(
    model_search: str, top_n_models: int = 10, task: Optional[str] = None
) -> Dict[str, Dict[str, str]]:
    """Display model data in a formatted table with dynamic sizing."""
    model_dict = get_model_data(model_search, top_n_models, task=task)

    # Return early if no models found
    if not model_dict:
        return model_dict

    # Calculate optimal widths based on content
    RANK_WIDTH, MODEL_ID_WIDTH, DETAIL_WIDTH, WIDTH = calculate_column_widths(model_dict)

    # Print header
    print("\n" + create_separator(WIDTH))
    header = f"Available Models for '{model_search}' (Top {top_n_models})"
    print(f"║{header.center(WIDTH-2)}║")
    print(create_separator(WIDTH, "="))

    # Print column headers
    print(format_table_row(["Rank", "Model ID", "Details"], [RANK_WIDTH, MODEL_ID_WIDTH, DETAIL_WIDTH]))
    print(create_separator(WIDTH, "-"))

    # Print each model's information
    for idx, (model_id, model_data) in enumerate(model_dict.items(), 1):
        # Print first row with rank and model ID
        print(format_table_row([str(idx), model_id, "Details:"], [RANK_WIDTH, MODEL_ID_WIDTH, DETAIL_WIDTH]))

        # Print model details
        for key, value in model_data.items():
            formatted_value = str(value).replace("\n", " ").strip()
            detail_text = f"{key}: {formatted_value}"
            print(format_table_row(["", "", detail_text], [RANK_WIDTH, MODEL_ID_WIDTH, DETAIL_WIDTH]))

        # Add separator between models
        if idx < len(model_dict):
            print(create_separator(WIDTH, "-"))

    # Print footer
    print(create_separator(WIDTH))
    print()  # Add final newline for cleaner output

    return model_dict


def format_key_value(key: str, value: str, indent: int = 0) -> str:
    """Format a key-value pair with proper indentation."""
    indent_str = " " * indent
    return f"{indent_str}{key}: {value}"



"""Define a custom logger class with colored output for logging to the console and/or a file and instantiate it"""


import logging
import os
from typing import Optional

import psutil

LEVELNAME_TO_LEVELVALUE = {
    "NOT SET": 0,
    "INFO": 20,
    "NOTICE": 25,
    "WARNING": 30,
}

VERBOSE_TO_LEVELNAME = {
    0: "NOT SET",
    1: "WARNING",
    2: "NOTICE",
    3: "INFO",
}

COLOR_CODES = {
    "DEBUG": "\033[1;36m",  # Bright Cyan
    "INFO": "\033[1;37m",  # Bright White
    "NOTICE": "\033[1;34m",  # Bright Blue
    "WARNING": "\033[1;33m",  # Bright Yellow
    "ERROR": "\033[1;31m",  # Bright Red
    "CRITICAL": "\033[1;35m",  # Bright Purple
}

RESET_CODE = "\033[0m"  # Resets to default terminal color


class ColoredFormatter(logging.Formatter):
    """Custom formatter that adds color to the log message"""

    def format(self, record):
        levelname = record.levelname
        if levelname in COLOR_CODES:
            color_code = COLOR_CODES[levelname]
            message = super().format(record)
            colored_message = f"{color_code}{message}{RESET_CODE}"
            return colored_message
        return super().format(record)


class EnhancedFormatter(logging.Formatter):
    """Enhanced formatter that adds CPU and memory usage to the log message"""

    def format(self, record):
        process = psutil.Process(os.getpid())
        record.cpu_usage = f"{psutil.cpu_percent(interval=None)}%"
        record.memory_usage = f"{psutil.virtual_memory().percent}%"
        record.process_mem = f"{process.memory_info().rss / (1024 ** 2):.2f} MB"
        return super().format(record)


class EnhancedColoredFormatter(ColoredFormatter, EnhancedFormatter):
    """Enhanced formatter that adds CPU and memory usage to the log message and colors the log message"""

# class Singleton(type):
#     _instances = {}
#     def __call__(cls, *args, **kwargs):
#         if cls not in cls._instances:
#             cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
#         return cls._instances[cls]

class ColoredLogger(logging.Logger):
    """Custom logger class with colored output for console and for logging to the console and/or a file"""

    def __init__(
        self,
        name: Optional[str] = None,
        verbose: int = 0,
        filename: Optional[str] = None,
        mode: str = "a+",
        log_cpu_and_memory_usage: bool = False,
    ):
        level_value = self._get_level_value_from_verbose(verbose)
        super().__init__(name, level_value)

        self.filename = filename
        self.mode = mode
        self.log_cpu_and_memory_usage = log_cpu_and_memory_usage

        self.add_custom_level(level_name="NOTICE", level_value=25)
        self.set_verbose(verbose)

    def info(self, msg: str, *args, **kwargs) -> None:
        if self.isEnabledFor(20):
            self._log(20, msg, args, **kwargs)

    def notice(self, msg: str, *args, **kwargs) -> None:
        if self.isEnabledFor(25):
            self._log(25, msg, args, **kwargs)

    def warning(self, msg: str, *args, **kwargs) -> None:
        if self.isEnabledFor(30):
            self._log(30, msg, args, **kwargs)

    def error(self, msg: str, *args, **kwargs) -> None:
        if self.isEnabledFor(40):
            self._log(40, msg, args, **kwargs)

    def warn_always(self, msg: str, *args, **kwargs) -> None:
        """Always log the message, regardless of the verbosity level"""
        _ = self.get_verbose()
        self.set_verbose(50)
        self._log(30, msg, args, **kwargs)
        self.set_verbose(_)

    def set_cpu_and_memory_usage(self, log_cpu_and_memory_usage: bool) -> None:
        """Set whether to additionaly log CPU and memory usage in the log message"""
        self.log_cpu_and_memory_usage = log_cpu_and_memory_usage
        self.formatter = self._get_formatter(self.name)
        self._infer_handler_from_verbose_and_add_it(self.verbose, self.level)

    def set_verbose(self, verbose: int) -> None:
        """Set the global verbosity level of the logger. If verbose is 0, the logger is disabled."""
        if verbose < 0:
            raise ValueError("verbose must be >= 0")

        self.verbose = verbose
        level_value = self._get_level_value_from_verbose(verbose)
        self.setLevel(level_value)
        self.formatter = self._get_formatter(self.name)
        self._infer_handler_from_verbose_and_add_it(verbose, level_value)

    def get_verbose(self) -> int:
        """Get the global verbosity level of the logger. If verbose is 0, the logger is disabled."""
        return self.verbose

    def set_log_file(self, filename: Optional[str], mode: str = "a+") -> None:
        """Set the log file to write to. If filename is None, the logger will write to the console."""
        if filename is not None:
            self.filename = filename
            self.mode = mode
            self._add_log_file_handler(self.level, self.filename, self.mode, self.formatter)

    def add_custom_level(self, level_name: str, level_value: int) -> None:
        """Add a custom log level to the logger"""
        if level_name not in logging._levelToName:
            setattr(logging, level_name, level_value)
            logging.addLevelName(level_value, level_name)

    def _get_formatter(self, name: str) -> logging.Formatter:
        date_fmt, fmt = self._get_formatting(name)
        if self.log_cpu_and_memory_usage:
            console_formatter = EnhancedColoredFormatter(fmt, datefmt=date_fmt)
            file_formatter = EnhancedFormatter(fmt, datefmt=date_fmt)
        else:
            console_formatter = ColoredFormatter(fmt, datefmt=date_fmt)
            file_formatter = logging.Formatter(fmt, datefmt=date_fmt)
        return console_formatter if self.filename is None else file_formatter

    def _get_formatting(self, name: Optional[str]) -> tuple:
        date_fmt = "%Y-%m-%d %H:%M:%S"
        if self.log_cpu_and_memory_usage:
            if name is None:
                fmt = "%(asctime)s - %(levelname)s - CPU: %(cpu_usage)s - Memory: %(memory_usage)s - Process Memory: %(process_mem)s - %(message)s"
            else:
                fmt = "%(asctime)s - %(name)s - %(levelname)s - CPU: %(cpu_usage)s - Memory: %(memory_usage)s - Process Memory: %(process_mem)s - %(message)s"
        else:
            if name is None:
                fmt = "%(asctime)s - %(levelname)s - %(message)s"
            else:
                fmt = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        return date_fmt, fmt

    def _get_level_value_from_verbose(self, verbose):
        level_name = VERBOSE_TO_LEVELNAME.get(abs(verbose), "INFO")
        level_value = LEVELNAME_TO_LEVELVALUE[level_name]
        return level_value

    def _add_console_handler(self, level: int, colored_formatter: ColoredFormatter) -> None:
        self.handlers.clear()
        ch = logging.StreamHandler()
        self._add_handler(level, colored_formatter, ch)

    def _add_log_file_handler(self, level: int, filename: str, mode: str, formatter: logging.Formatter) -> None:
        self.handlers.clear()
        fh = logging.FileHandler(filename=filename, mode=mode)
        self._add_handler(level, formatter, fh)

    def _add_handler(self, level: int, formatter: logging.Formatter, ch: logging.Handler):
        ch.setLevel(level)
        ch.setFormatter(formatter)
        self.addHandler(ch)

    def _infer_handler_from_verbose_and_add_it(self, verbose, level_value):
        if verbose != 0:
            self.disabled = False
            if self.filename is None:
                self._add_console_handler(level_value, self.formatter)
            else:
                self._add_log_file_handler(level_value, self.filename, self.mode, self.formatter)
        else:
            self.disabled = True
            self.handlers.clear()

logger = ColoredLogger(verbose=3)
